/*
****************************************************************************
****************************************************************************
**                                                                        **
**                XXXX     X    X         X XXXXXX X                      **
**               X    X   X X   X    X    X X      X   X                  **
**               X       X   X  X   X X   X X      X  X                   **
**               X   XX XXXXXXX X  X   X  X XXXX   X X                    **
**               X    X X     X X X     X X X      XX                     **
**               X    X X     X XX       XX X      X                      **
**                XXXX  X     X X         X XXXXXX XXXXXX                 **
**                                                                        **
**                                                                        **
**                           BIALYSTOK 2018                               **
**                                                                        **
**                          MODULE TCPExample                             **
**                                                                        **
**                     Mikrokontroler ARM: LPC2378                        **
**                                                                        **
**                     Wersja z dnia : 2021.04.12                         **
**                     Napisal : Andrzej Pawluczuk                        **
**                                                                        **
****************************************************************************
****************************************************************************
*/

#include <LPC23xx.H>
#include "system.h"
#include "types.h"
#include "systime.h"
#include "irqvectors.h"
#include "serialconst.h"
#include "uart0.h"
#include "strings.h"
#include "emacdriver.h"
#include "ipservice.h"
#include "tcpservice.h"
#include "trace.h"

#define IRQCallCntLimit                 100

#define Cr                              0x0D

typedef struct {
                 EMACInstanceRecT *          EMACLayer ;
                 IPLayerInstanceRecT *       IPLayer ;
                 TCPLayerInstanceRecT *      TCPLayer ;
                 PtrTCPSocketDescriptionRecT TCPSocket ;
                 PtrTCPSocketDescriptionRecT TelnetSocket ;
                 USHORT                      ActiveConfig ;
                 UCHAR                       UserName [ 20 ] ;
                 UCHAR                       Password [ 20 ] ;
               } TCPSeverInstanceRecT ;
typedef TCPSeverInstanceRecT * PtrTCPSeverInstanceRecT ;

static TCPSeverInstanceRecT TCPSeverInstance ;
static ULONG IRQCallCounter ;

#define MachineIPAddress        0xC0A80037
//                              ^^^ 192.168.0.55
#define MachineIPSubnet         0xFFFFFF00
//                              ^^^ 255.255.255.0
#define MachineGWIPAddress      0xC0A800FE
//                              ^^^ 192.168.0.254
#define TCPPortNumber           2000
#define TelnetPortNumber        23


UCHAR MachineMACAddr [ 6 ]     = { 0x08 , 0x14 , 0x16 , 0x22 , 0x28 , 0x2D } ;




static __irq void T1_IRQHandler ( void )
{
  /*-------------------------------------------------------------------------*/
  IRQCallCounter ++ ;
  if ( IRQCallCounter >= IRQCallCntLimit )
  {
    IRQCallCounter = 0 ;
  } /* if */ ;
  T1IR = 1 ;
  VICVectAddr = 0 ;
} /* T1_IRQHandler */


static void HardwareInit ( void ) 
{
  /*-------------------------------------------------------------------------*/
  PINSEL10 = 0 ;
  T1MR0 = ( PCLK / 1000 ) - 1 ;
  T1MCR = 3 ;
  T1TCR = 1 ;
  VICVectAddr5 = ( unsigned long ) T1_IRQHandler ;
  VICVectCntl5 = 15 ;
  VICIntEnable = ( 1 << TIMER1_Interrupt ) ;
  UART0InitEnvir ( ) ;
} /* HardwareInit */


static void StartNet ( PtrTCPSeverInstanceRecT Instance )
{
  EtherSpeedMode EthMode ;
  UCHAR LocalMyMACAddr [ 6 ] ;
  USHORT Loop ;
  /*-------------------------------------------------------------------------*/
  for ( Loop = 0 ; Loop < 6 ; Loop ++ )
    LocalMyMACAddr [ Loop ] = MachineMACAddr [ Loop ] ;
  EthMode = Mode_AutoNeg ;
//  EthMode = Mode_10MBIT ;
//  EthMode = Mode_100MBIT ;
  Instance -> EMACLayer = InitPHYLayerInstance ( LocalMyMACAddr , EthMode ) ;
  Instance -> IPLayer = InitIPLayerInstance ( Instance -> EMACLayer , MachineIPAddress ,
                                              MachineIPSubnet ) ;
  Instance -> IPLayer -> MyDefaultGWIPAddress = MachineGWIPAddress ;
  Instance -> TCPLayer = StandardAddTCPInstance ( Instance -> IPLayer ) ;
} /* StartNet */


UCHAR Hello1 [ ]      = "\r\nHello: Serwer polaczen TCP\r\n" ;
UCHAR Hello2 [ ]      = "\r\nMoj adres IP=" ;
UCHAR Hello3 [ ]      = "\r\nMaska podsieci=" ;
UCHAR Hello4 [ ]      = "\r\nAdres bramy domyslnej=" ;
UCHAR Hello5 [ ]      = "\r\nNasluch na porcie TCP=" ;
UCHAR Hello6 [ ]      = "\r\nAdres MAC=" ;
UCHAR Hello7 [ ]      = "\r\n************************\r\n" ;
UCHAR Error1 [ ]      = "\r\nProblem z siecia: " ;

static void HelloMessage ( void )
{
  /*-------------------------------------------------------------------------*/
  UART0SendString ( Hello1 ) ;
  UART0SendString ( Hello2 ) ;
  TraceIP ( MachineIPAddress ) ;
  UART0SendString ( Hello3 ) ;
  TraceIP ( MachineIPSubnet ) ;
  UART0SendString ( Hello4 ) ;
  TraceIP ( MachineGWIPAddress ) ;
  UART0SendString ( Hello5 ) ;
  TraceUSHORTdec ( TCPPortNumber ) ;
  UART0SendString ( Hello6 ) ;
  TraceMAC ( MachineMACAddr ) ;
  UART0SendString ( Hello7 ) ;
} /* HelloMessage */


USHORT TCPConnectOperation ( PtrTCPSocketDescriptionRecT TCPSocket ,
                             ULONG                       RemoteAddress ,
                             USHORT                      RemotePort )
{
  /*-------------------------------------------------------------------------*/
  UART0SendString ( "Proba nawiazania polaczenia z IP=" ) ;
  TraceIP ( RemoteAddress ) ;
  UART0SendString ( " port=" ) ;
  TraceUSHORTdec ( RemotePort ) ;
  UART0SendString ( "\r\n" ) ;
  return ( TRUE ) ;
} /* TCPConnectOperation */


void TCPEstablishOperation ( PtrTCPSocketDescriptionRecT TCPSocket )
{
  /*-------------------------------------------------------------------------*/
  UART0SendString ( "Polaczenie zestawione\r\n" ) ;
} /* TCPEstablishOperation */

                                                                                
void TCPCloseOperation ( PtrTCPSocketDescriptionRecT TCPSocket )
{
  /*-------------------------------------------------------------------------*/
  UART0SendString ( "Polaczenie zamkniete\r\n" ) ;
} /* TCPCloseOperation */


void TCPRecvOperation ( PtrTCPSocketDescriptionRecT TCPSocket ,
                        ULONG                       SenderIPAddress ,
                        USHORT                      SenderPortNo ,
                        ULONG                       DestIPAddress ,
                        USHORT                      DestPortNo ,
                        void *                      DataBuffer ,
                        USHORT                      DataBufferSize )
{
  UCHAR * Buffer ;
  USHORT Loop ;
  UCHAR SendData [ 32 ] ;
  UCHAR * ChPtr ;
  /*-------------------------------------------------------------------------*/
  if ( DataBufferSize > 0 )
  {
    Buffer = ( UCHAR * ) DataBuffer ;
    UART0SendString ( "Przyszly dane: " ) ;
    for ( Loop = 0 ; Loop < DataBufferSize ; Loop ++ )
      UART0Send ( * ( Buffer  + Loop ) ) ;
    UART0Send ( '\r' ) ;
    UART0Send ( '\n' ) ;
    if ( DataBufferSize > 30 )
      DataBufferSize = 30 ;
    ChPtr = SendData ;
    * ChPtr ++ = '*' ;
    for ( Loop = 0 ; Loop < DataBufferSize ; Loop ++ )
      * ChPtr ++ = * ( Buffer  + Loop ) ;
    * ChPtr ++ = '*' ;
    TCPSendData ( TCPSocket , SendData , DataBufferSize + 2 ) ;
  } /* if */ ;
} /* TCPRecvOperation */


USHORT TelnetConnectOperation ( PtrTCPSocketDescriptionRecT TCPSocket ,
                                ULONG                       RemoteAddress ,
                                USHORT                      RemotePort )
{
  /*-------------------------------------------------------------------------*/
  if ( TCPSeverInstance . ActiveConfig )
    return ( FALSE ) ;
  TCPSeverInstance . ActiveConfig = 1 ;
  return ( TRUE ) ;
} /* TelnetConnectOperation */


UCHAR LoginText [ ] = "Uzytkownik:" ;
UCHAR PasswordText [ ] = "\r\nHaslo:" ;
UCHAR LoginFailText [ ] = "\r\nBlad autoryzacji.\r\n" ;
UCHAR MenuText [ ] = "\r\nOpcje konfiguracyjne:\r\n  1 - Opcja 1\r\n  2 - Opcja 2\r\n  3 - Opcja 3\r\n  0 - Zakoncz\r\n" ;
UCHAR Option0ExecuteText [ ] = "\r\nZakonczenie polaczenia.\r\n\r\n" ;
UCHAR Option1ExecuteText [ ] = "\r\nRealizacja opcji 1" ;
UCHAR Option2ExecuteText [ ] = "\r\nRealizacja opcji 2" ;
UCHAR Option3ExecuteText [ ] = "\r\nRealizacja opcji 3" ;

#define OptionPrefix                0xFF
#define WILLOptionCode              0xFB
#define WONTOptionCode              0xFC
#define DOOptionCode                0xFD
#define DONTOptionCode              0xFE

#define EchoOptionCode	            0x01
//#define SupressGoAhead					    0x03

UCHAR * AnalyseOption ( UCHAR *                     TelnetBuffer ,
                        USHORT *                    TelNetBufferSize ,
                        PtrTCPSocketDescriptionRecT TCPSocket )
{
  UCHAR OptionCode ;
	UCHAR OptionValue ;
	UCHAR OptionBuffer [ 30 ] ;
	UCHAR Index ;
	UCHAR ReplyCode ;
  // FF FB 1F FF FB 20 FF FB 18 FF FB 27 FF FD 01 FF FB 03 FF FD 03 
  /*-------------------------------------------------------------------------*/
	Index = 0 ;
  while ( ( * TelnetBuffer ) == OptionPrefix )
  {
    TelnetBuffer ++ ;
    ( * TelNetBufferSize ) -- ;
    OptionCode = * TelnetBuffer ++ ;
    ( * TelNetBufferSize ) -- ;
	  OptionValue = * TelnetBuffer ++ ;
    ( * TelNetBufferSize ) -- ;
	  ReplyCode = 0 ;
    switch ( OptionCode )
    {
      case WILLOptionCode              :
        break ;
      case WONTOptionCode              :
				ReplyCode = DONTOptionCode ;
        break ;
      case DOOptionCode                :
        if ( OptionValue == EchoOptionCode )
		  		ReplyCode = WILLOptionCode ;
				else
		  		ReplyCode = WONTOptionCode ;
        break ;
      case DONTOptionCode              :
				ReplyCode = WONTOptionCode ;
        break ;
    } /* switch */ ;
		if ( ReplyCode )
		{
			OptionBuffer [ Index ++ ] = OptionPrefix ;
			OptionBuffer [ Index ++ ] = ReplyCode ;
			OptionBuffer [ Index ++ ] = OptionValue ;
		} /* if */ ;
  } /* while */ ;
  if ( Index )
    TCPSendData ( TCPSocket , OptionBuffer , Index ) ;
  return TelnetBuffer ;
} /* AnalyseOption */


USHORT UserNameIsValid ( UCHAR * UserName )
{
  /*-------------------------------------------------------------------------*/
  if ( CompareString ( UserName , "USER" , 4 ) )
    return 1 ;
  else
    return 0 ;
} /* TelnetEstablishOperation */


USHORT PasswordIsValid ( UCHAR * PassW )
{
  /*-------------------------------------------------------------------------*/
  if ( CompareString ( PassW , "PASSW" , 4 ) )
    return 1 ;
  else
    return 0 ;
} /* PasswordIsValid */


static void SendEchoOffReq ( PtrTCPSocketDescriptionRecT TCPSocket )
{
	UCHAR Option [ 3 ] ;
  /*-------------------------------------------------------------------------*/
	Option [ 0 ] = OptionPrefix ;
	Option [ 1 ] = DONTOptionCode ;
	Option [ 2 ] = EchoOptionCode ;
  TCPSendData ( TCPSocket , Option , 3 ) ;
} /* SendEchoOffReq */


static void SendEchoOnReq ( PtrTCPSocketDescriptionRecT TCPSocket )
{
	UCHAR Option [ 3 ] ;
  /*-------------------------------------------------------------------------*/
	Option [ 0 ] = OptionPrefix ;
	Option [ 1 ] = DOOptionCode ;
	Option [ 2 ] = EchoOptionCode ;
  TCPSendData ( TCPSocket , Option , 3 ) ;
} /*SendEchoOnReq */


void TelnetEstablishOperation ( PtrTCPSocketDescriptionRecT TCPSocket )
{
  /*-------------------------------------------------------------------------*/
  SendEchoOffReq ( TCPSocket ) ;
  TCPSendData ( TCPSocket , LoginText , StrLen ( LoginText ) ) ;
  TCPSeverInstance . UserName [ 0 ] = 0 ;
  TCPSeverInstance . Password [ 0 ] = 0 ;
} /* TelnetEstablishOperation */

                                                                                
USHORT ExtractUserName ( UCHAR * TextBuffer , 
                         USHORT  DataBufferSize )
{
  UCHAR Ch ;
  /*-------------------------------------------------------------------------*/
  while ( DataBufferSize )
  {
    Ch = * TextBuffer ++ ;
    DataBufferSize -- ;
    if ( Ch == Cr )
      return TRUE ;
    AppendCh ( TCPSeverInstance . UserName , Ch ) ;
  } /* while */ ;
  return FALSE ;
} /* ExtractUserName */


USHORT ExtractPassword ( UCHAR * TextBuffer , 
                         USHORT  DataBufferSize )
{
  UCHAR Ch ;
  /*-------------------------------------------------------------------------*/
  while ( DataBufferSize )
  {
    Ch = * TextBuffer ++ ;
    DataBufferSize -- ;
    if ( Ch == Cr )
      return TRUE ;
    AppendCh ( TCPSeverInstance . Password , Ch ) ;
  } /* while */ ;
  return FALSE ;
} /* ExtractPassword */


void TelnetRecvUserNameOperation ( PtrTCPSocketDescriptionRecT TCPSocket ,
                                   ULONG                       SenderIPAddress ,
                                   USHORT                      SenderPortNo ,
                                   ULONG                       DestIPAddress ,
                                   USHORT                      DestPortNo ,
                                   void *                      DataBuffer ,
                                   USHORT                      DataBufferSize ) ;
                                   
void TelnetCloseOperation ( PtrTCPSocketDescriptionRecT TCPSocket )
{
  /*-------------------------------------------------------------------------*/
  TCPSeverInstance . ActiveConfig = 0 ;
  TCPSetNewRecOperation ( TCPSocket , TelnetRecvUserNameOperation ) ;
} /* TelnetCloseOperation */


void TelnetRecvOptionOperation ( PtrTCPSocketDescriptionRecT TCPSocket ,
                                 ULONG                       SenderIPAddress ,
                                 USHORT                      SenderPortNo ,
                                 ULONG                       DestIPAddress ,
                                 USHORT                      DestPortNo ,
                                 void *                      DataBuffer ,
                                 USHORT                      DataBufferSize )
{
  UCHAR * TextBuffer ;
  /*-------------------------------------------------------------------------*/  
  if ( DataBufferSize > 0 )
  {
    TextBuffer = AnalyseOption ( ( UCHAR * ) DataBuffer , & DataBufferSize , TCPSocket ) ;
    if ( DataBufferSize == 0 )
      return ;
    switch ( * TextBuffer )
    {
      case '0' :
        TCPSendData ( TCPSocket , Option0ExecuteText , StrLen ( Option0ExecuteText ) ) ;
        TCPSetNewRecOperation ( TCPSocket , TelnetRecvOptionOperation ) ;
        TCPClose ( TCPSocket ) ;
        TCPSeverInstance . ActiveConfig = 0 ;
        break ;
      case '1' :
        TCPSendData ( TCPSocket , Option1ExecuteText , StrLen ( Option1ExecuteText ) ) ;
        TCPSendData ( TCPSocket , MenuText , StrLen ( MenuText ) ) ;
        break ;
      case '2' :
        TCPSendData ( TCPSocket , Option2ExecuteText , StrLen ( Option2ExecuteText ) ) ;
        TCPSendData ( TCPSocket , MenuText , StrLen ( MenuText ) ) ;
        break ;
      case '3' :
        TCPSendData ( TCPSocket , Option3ExecuteText , StrLen ( Option3ExecuteText ) ) ;
        TCPSendData ( TCPSocket , MenuText , StrLen ( MenuText ) ) ;
        break ;
    } /* switch */ ;
  } /* if */ ;
} /* TelnetRecvOptionOperation */


void TelnetRecvPasswordOperation ( PtrTCPSocketDescriptionRecT TCPSocket ,
                                   ULONG                       SenderIPAddress ,
                                   USHORT                      SenderPortNo ,
                                   ULONG                       DestIPAddress ,
                                   USHORT                      DestPortNo ,
                                   void *                      DataBuffer ,
                                   USHORT                      DataBufferSize )
{
  UCHAR * TextBuffer ;
  USHORT Ready ;
  /*-------------------------------------------------------------------------*/
  if ( DataBufferSize > 0 )
  {
    TextBuffer = AnalyseOption ( ( UCHAR * ) DataBuffer , & DataBufferSize , TCPSocket ) ;
    if ( DataBufferSize == 0 )
      return ;
    Ready = ExtractPassword ( TextBuffer , DataBufferSize ) ;
    if ( ! Ready )
      return ;
    if ( PasswordIsValid ( TCPSeverInstance . Password ) )
    {
      SendEchoOnReq ( TCPSocket ) ;
      TCPSendData ( TCPSocket , MenuText , StrLen ( MenuText ) ) ;
      TCPSetNewRecOperation ( TCPSocket , TelnetRecvOptionOperation ) ;
    } /* if ... */
    else
    {
      TCPSendData ( TCPSocket , LoginFailText , StrLen ( LoginFailText ) ) ;
      TCPClose ( TCPSocket ) ;
      TCPSeverInstance . ActiveConfig = 0 ;
    } /* if ... else */ ;
  } /* if */ ;
} /* TelnetRecvPasswordOperation */


void TelnetRecvUserNameOperation ( PtrTCPSocketDescriptionRecT TCPSocket ,
                                   ULONG                       SenderIPAddress ,
                                   USHORT                      SenderPortNo ,
                                   ULONG                       DestIPAddress ,
                                   USHORT                      DestPortNo ,
                                   void *                      DataBuffer ,
                                   USHORT                      DataBufferSize )
{
  UCHAR * TextBuffer ;
  USHORT Ready ;
  /*-------------------------------------------------------------------------*/
  if ( DataBufferSize > 0 )
  {
    TextBuffer = AnalyseOption ( ( UCHAR * ) DataBuffer , & DataBufferSize , TCPSocket ) ;
    if ( DataBufferSize == 0 )
    {
      return ;
		} /* if */ ;
    Ready = ExtractUserName ( TextBuffer , DataBufferSize ) ;
    if ( ! Ready )
      return ;
    if ( UserNameIsValid ( TCPSeverInstance . UserName ) )
    {
      TCPSendData ( TCPSocket , PasswordText , StrLen ( PasswordText ) ) ;
      TCPSetNewRecOperation ( TCPSocket , TelnetRecvPasswordOperation ) ;
    } /* if ... */
    else
    {
      TCPSendData ( TCPSocket , LoginFailText , StrLen ( LoginFailText ) ) ;
      TCPClose ( TCPSocket ) ;
      TCPSeverInstance . ActiveConfig = 0 ;
    } /* if ... else */ ;
  } /* if */ ;
} /* TelnetRecvUserNameOperation */


void SoftwareInit ( void )
{
  /*-------------------------------------------------------------------------*/
  IRQCallCounter = 0 ;
  TCPSeverInstance . EMACLayer = NIL ;
  TCPSeverInstance . IPLayer = NIL ;
  TCPSeverInstance . TCPLayer = NIL ;
  TCPSeverInstance . TCPSocket = NIL ;
  TCPSeverInstance . TelnetSocket = NIL ;
  TCPSeverInstance . ActiveConfig = 0 ;
} /* SoftwareInit */


static void Halt ( UCHAR * Message ,
                   UCHAR   Selector )
{
  /*-------------------------------------------------------------------------*/
  UART0SendString ( Message ) ;
  UART0Send ( Selector ) ;
  UART0Send ( 0x0D ) ;
  UART0Send ( 0x0A ) ;
  for ( ; ; )
  {
    NOP ( ) ;
  } /* loop */ ;
} /* Halt */


int main ( void )
{
  /*-------------------------------------------------------------------------*/
  SysInit ( ) ;
  InitSysTime ( ) ;
  HardwareInit ( )  ;
  SoftwareInit ( ) ;
  UART0HardwInit( FOSC , N81Mode , 9600 ) ;
  HelloMessage ( ) ;
  StartNet ( & TCPSeverInstance ) ;
  TCPSeverInstance . TCPSocket = CreateTCPSocket ( TCPSeverInstance . TCPLayer ) ;
  if ( BindTCPSocket ( TCPSeverInstance . TCPSocket , TCPPortNumber ) != TCP_BindPortOK )
  {
    Halt ( Error1 , 'A' ) ;
  } /* if */ ;
  if ( TCPListen ( TCPSeverInstance . TCPSocket , TCPConnectOperation , TCPEstablishOperation ,
                   TCPRecvOperation , TCPCloseOperation ) != TCP_ListenPortOK )
  {
    Halt ( Error1 , 'B' ) ;
  } /* if */ ;
  TCPSeverInstance . TelnetSocket = CreateTCPSocket ( TCPSeverInstance . TCPLayer ) ;
  if ( ! TCPSeverInstance . TelnetSocket )
  {
    Halt ( Error1 , 'C' ) ;
  } /* if */ ;
  if ( BindTCPSocket ( TCPSeverInstance . TelnetSocket , TelnetPortNumber ) != TCP_BindPortOK )
  {
    Halt ( Error1 , 'D' ) ;
  } /* if */ ;
  if ( TCPListen ( TCPSeverInstance . TelnetSocket , TelnetConnectOperation , TelnetEstablishOperation ,
                   TelnetRecvUserNameOperation , TelnetCloseOperation ) != TCP_ListenPortOK )
  {
    Halt ( Error1 , 'E' ) ;
  } /* if */ ;
  for ( ; ; )
  {
    SysTimerPool ( ) ;
    TCPSeverInstance . IPLayer -> IPLayerPoolService ( TCPSeverInstance . IPLayer ) ;
  } /* loop */ ;
} /* main */

